﻿using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;

namespace HeuristicLab.EvolutionaryTracking {
  [Item("SymbolicExpressionTreePruningAnalyzer", "An analyzer that identifies introns in the population and optionally removes them.")]
  [StorableClass]
  public sealed class SymbolicExpressionTreePruningAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string SymbolicExpressionTreeQualityParameterName = "Quality";
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string MatchVariablesParameterName = "MatchVariableNames";
    private const string MatchVariableWeightsParameterName = "MatchVariableWeights";
    private const string MatchConstantValuesParameterName = "MatchConstantValues";
    private const string ResultsParameterName = "Results";
    private const string SymbolicExpressionInterpreterParameterName = "SymbolicExpressionTreeInterpreter";
    private const string SymbolicRegressionProblemDataParameterName = "ProblemData";
    private const string FrequencyThresholdParameterName = "FrequencyThreshold";
    private const string IgnoreSubtreeFrequenciesParameterName = "IgnoreFrequencies";

    [Storable]
    private SymbolicRegressionSolutionImpactValuesCalculator calculator;
    [Storable]
    private SymbolicExpressionTreeNodeSimilarityComparer comparer;

    #region Parameters
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public ValueParameter<BoolValue> MatchVariableNamesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariablesParameterName]; }
    }
    public ValueParameter<BoolValue> MatchVariableWeightsParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariableWeightsParameterName]; }
    }
    public ValueParameter<BoolValue> MatchConstantValuesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchConstantValuesParameterName]; }
    }
    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> SymbolicExpressionTreeQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters[SymbolicExpressionTreeQualityParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter> SymbolicExpressionTreeInterpreterParameter {
      get { return (ILookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>)Parameters[SymbolicExpressionInterpreterParameterName]; }
    }
    public ILookupParameter<RegressionProblemData> SymbolicRegressionProblemDataParameter {
      get { return (ILookupParameter<RegressionProblemData>)Parameters[SymbolicRegressionProblemDataParameterName]; }
    }
    public ValueParameter<DoubleValue> FrequencyThresholdParameter {
      get { return (ValueParameter<DoubleValue>)Parameters[FrequencyThresholdParameterName]; }
    }
    public ValueParameter<BoolValue> IgnoreSubtreeFrequenciesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[IgnoreSubtreeFrequenciesParameterName]; }
    }
    #endregion

    #region Parameter properties
    public bool MatchValues {
      get { return MatchConstantValuesParameter.Value.Value; }
    }
    public bool MatchWeights {
      get { return MatchVariableWeightsParameter.Value.Value; }
    }
    public bool MatchVariables {
      get { return MatchVariableNamesParameter.Value.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public SymbolicDataAnalysisExpressionTreeInterpreter SymbolicExpressionTreeInterpreter {
      get { return SymbolicExpressionTreeInterpreterParameter.ActualValue; }
    }
    public RegressionProblemData ProblemData {
      get { return SymbolicRegressionProblemDataParameter.ActualValue; }
    }
    public DoubleValue FrequencyThreshold {
      get { return FrequencyThresholdParameter.Value; }
    }
    public bool IgnoreSubtreesFrequencies {
      get { return IgnoreSubtreeFrequenciesParameter.Value.Value; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreePruningAnalyzer(bool deserializing) : base(deserializing) { }
    private SymbolicExpressionTreePruningAnalyzer(SymbolicExpressionTreePruningAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      calculator = original.calculator;
      comparer = original.comparer;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreePruningAnalyzer(this, cloner);
    }

    public SymbolicExpressionTreePruningAnalyzer() {
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>(SymbolicExpressionTreeQualityParameterName, "The qualities of the symbolic expression trees"));
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariablesParameterName, "Specify if the symbolic expression tree comparer should match variable names.", new BoolValue(true)));
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariableWeightsParameterName, "Specify if the symbolic expression tree comparer should match variable weights.", new BoolValue(true)));
      Parameters.Add(new ValueParameter<BoolValue>(MatchConstantValuesParameterName, "Specify if the symbolic expression tree comparer should match constant values.", new BoolValue(true)));
      Parameters.Add(new LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>(SymbolicExpressionInterpreterParameterName, "Interpreter for symbolic expression trees"));
      Parameters.Add(new LookupParameter<RegressionProblemData>(SymbolicRegressionProblemDataParameterName, "The symbolic data analysis problem."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new ValueParameter<DoubleValue>(FrequencyThresholdParameterName, "The frequency threshold a fragment has to exceed before being removed."));
      Parameters.Add(new ValueParameter<BoolValue>(IgnoreSubtreeFrequenciesParameterName, "Specifies whether or not subtree frequencies should be taken into account."));

      calculator = new SymbolicRegressionSolutionImpactValuesCalculator();
      comparer = new SymbolicExpressionTreeNodeSimilarityComparer { MatchConstantValues = MatchValues, MatchVariableWeights = MatchWeights, MatchVariableNames = MatchVariables };

      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }

    public override IOperation Apply() {
      UpdateCounter.Value++;
      if (UpdateCounter.Value != UpdateInterval.Value) return base.Apply();
      UpdateCounter.Value = 0;

      //var prunedSubtrees = new IntValue(0);
      int prunedNodes = 0;
      int prunedTrees = 0;
      DataTable prunedSubtreesTable;
      if (Results.ContainsKey("Pruning")) {
        prunedSubtreesTable = (DataTable)Results["Pruning"].Value;
      } else {
        prunedSubtreesTable = new DataTable("Pruning");
        var prunedNodesRow = new DataRow("Pruned Nodes") { VisualProperties = { StartIndexZero = true } };
        var prunedTreesRow = new DataRow("Pruned Trees") { VisualProperties = { StartIndexZero = true } };
        prunedSubtreesTable.Rows.Add(prunedNodesRow);
        prunedSubtreesTable.Rows.Add(prunedTreesRow);
        Results.Add(new Result("Pruning", prunedSubtreesTable));
      }

      var trees = SymbolicExpressionTreeParameter.ActualValue.ToList();
      var qualities = SymbolicExpressionTreeQualityParameter.ActualValue.ToList();
      var prunedTreesMap = new bool[trees.Count];

      for (int i = 0; i != trees.Count; ++i) {
        var model = new SymbolicRegressionModel(trees[i], SymbolicExpressionTreeInterpreter);
        var root = trees[i].Root.GetSubtree(0).GetSubtree(0); // skip the program root symbol and the start symbol
        var nodes = root.IterateNodesPrefix().Select((node, index) => new { Node = node, Index = index }).ToList();
        for (int j = 0; j < nodes.Count; ++j) {
          var n = nodes[j];
          if (n.Node is ConstantTreeNode) continue;
          double impact = calculator.CalculateImpactValue(model, n.Node, ProblemData, ProblemData.TrainingIndices, qualities[i].Value);

          //if (impact.IsAlmost(0.0)) {
          if (impact < 0.001) {
            bool prune = true;
            if (!IgnoreSubtreesFrequencies) {
              double frequency = (double)trees.Count(t => t.Root.ContainsFragment(new Fragment { Root = n.Node, Index = n.Index }, comparer)) / trees.Count;
              // remove node if its frequency is above the threshold
              if (frequency < FrequencyThreshold.Value) { prune = false; }
            }
            if (!prune) continue;
            var parent = n.Node.Parent;
            if (parent != null && !(parent.Symbol is StartSymbol)) {
              double replacementValue = calculator.CalculateReplacementValue(model, n.Node, ProblemData, ProblemData.TrainingIndices);
              var constantNode = new ConstantTreeNode(new Constant()) { Value = replacementValue };
              var index = parent.IndexOfSubtree(n.Node);
              parent.RemoveSubtree(index);
              parent.InsertSubtree(index, constantNode);
              // skip the removed subtree
              int length = n.Node.GetLength();
              j += length;

              if (!prunedTreesMap[i]) {
                prunedTrees++;
                prunedTreesMap[i] = true;
              }
              prunedNodes += (length - 1);
            }
          }
        }
      }

      prunedSubtreesTable.Rows["Pruned Nodes"].Values.Add(prunedNodes);
      prunedSubtreesTable.Rows["Pruned Trees"].Values.Add(prunedTrees);

      return base.Apply();
    }

    public bool EnabledByDefault { get { return true; } }

  }
}
